# RA, 2020-06-17

import typing


def kplets(k: int, s: str, no_remainder=True) -> typing.Iterable[str]:
    """
    When k = 3, from a string like s = 'GCATCGAGC'
    generate the triples 'GCA', 'TCG', 'AGC'.
    Similarly for other k >= 1.
    The string must have length divisible by k
    when no_remainder is True; otherwise
    the trailing characters are ignored.
    """
    assert (k >= 1)

    if no_remainder and (len(s) % k):
        raise ValueError(F"The string should have length a multiple of {k}")

    for n in range(0, (len(s) // k) * k, k):
        yield s[n:(n + k)]


def triplets(s: str) -> typing.Iterable[str]:
    """
    From a string like s = 'GCATCGAGC'
    generate the triples 'GCA', 'TCG', 'AGC'.
    """
    yield from kplets(3, s, no_remainder=True)


def reverse(s: str) -> str:
    """
    For a string s return the reversed string s[::-1].
    """
    return s[::-1]


def lines(s: str) -> typing.Iterable[str]:
    """
    Split the string at '\n' and yield the lines.
    Applies str.strip to each line.
    """
    yield from map(str.strip, s.strip().split('\n'))


def nnna(s: str, n=3, m=1) -> typing.Iterable[typing.Tuple[str, str]]:
    """
    Takes a string like
        UUU F UUC F ...
    where white spaces and new lines are ignored.

    Generates tuples
        ('UUU', 'F'), ('UUC', 'F'), ...
    """
    assert (n >= 0) and (m >= 0) and (n + m >= 1)

    import re
    s = re.sub(r"\s+", "", s)

    if (len(s) % (n + m)):
        raise ValueError(F"The string's length {len(s)} (ignoring whitespace) is not a multiple of (n + m) = {n + m}.")

    from tcga.utils import First
    f = First(lambda x: kplets(n + m, x)).each(lambda x: (x[0:n], x[n:(n + m)]))

    yield from f(s)



# RA, 2020-06-24


import io
import pandas
import hashlib
import base64
import json
import contextlib
import pathlib
import zipfile
import typing
import datetime
import inspect
import copy
import tcga.refs

import shutil
import urllib.request


class _:
    downloads = []

    class Download:
        def __init__(self, **kwargs):
            self._ = pandas.Series({
                'meta': {}, 'url': None, 'local_folder': None, 'again': False,
                **(kwargs or {}),
            })

        @property
        def local_folder(self) -> pathlib.Path:
            return self._.local_folder

        def to(self, *, rel_path=None, abs_path=None):
            if (bool(abs_path) == bool(rel_path)):
                raise RuntimeError("Either rel_path /or/ abs_path must be specified.")
            if abs_path:
                local_folder = pathlib.Path(abs_path)
            else:
                f = inspect.currentframe().f_back
                local_folder = (pathlib.Path(f.f_code.co_filename).parent / rel_path)
            return type(self)(**{**self._, 'local_folder': local_folder})

        def __call__(self, url: str):
            return type(self)(**{**self._, 'url': url})

        def again(self, again=False):
            return type(self)(**{**self._, 'again': again})

        def with_meta(self, meta: dict):
            assert 'source' not in meta
            assert 'datetime' not in meta
            assert json.loads(json.dumps(meta)) == meta, "The specified metadata should be json-stable."
            return type(self)(**{**self._, 'meta': meta})

        @property
        def now(self):
            if not self._.url:
                raise RuntimeError("What do you want to download?")
            if not self._.local_folder:
                raise RuntimeError("Use  .to(...)  to set the local folder before downloading.")

            name = base64.urlsafe_b64encode(hashlib.sha256(self._.url.encode()).digest()).decode()
            local_file = (pathlib.Path(self._.local_folder) / name).with_suffix(".zip")

            if self._.again or not local_file.exists():
                local_file.parent.mkdir(exist_ok=True, parents=True)

                assert 'url' not in self._.meta
                assert 'datetime' not in self._.meta

                with contextlib.closing(urllib.request.urlopen(url=self._.url)) as rd:
                    with zipfile.ZipFile(local_file, mode='w', compression=zipfile.ZIP_DEFLATED, compresslevel=9) as zf:
                        with zf.open("data", mode='w', force_zip64=True) as fd:
                            shutil.copyfileobj(rd, fd)
                        with zf.open("meta", mode='w') as fd:
                            meta = self._.meta.copy()
                            meta['source'] = self._.url
                            meta['datetime'] = datetime.datetime.now(tz=datetime.timezone.utc).isoformat(sep=' ')
                            fd.write(json.dumps(meta).encode())

            assert local_file.exists()

            with zipfile.ZipFile(local_file, mode='r') as zf:
                with zf.open("meta") as fd:
                    meta = json.loads(fd.read().decode())
                    for (k, v) in self._.meta.items():
                        if k not in meta:
                            raise RuntimeWarning("The specified metadata is not in the file. Use again(True).")
                        if not (meta[k] == v):
                            raise RuntimeWarning(
                                "The specified metadata differs from the file. Use again(True).")
                    meta = {**self._.meta, **meta}

            class Local:
                def __init__(self, local_file: pathlib.Path, meta: dict):
                    self.local_file = local_file
                    self.meta = meta

                def __call__(self):
                    return self

                @contextlib.contextmanager
                def open(self, mode='r'):
                    assert mode in ["r", "rb"]
                    with zipfile.ZipFile(self.local_file, mode='r') as _:
                        with _.open("data") as _:
                            if mode == "rb":
                                yield _
                            else:
                                yield io.TextIOWrapper(_)

                @property
                def bytes(self) -> bytes:
                    with self.open(mode='rb') as _:
                        return _.read()

                @property
                def text(self) -> str:
                    with self.open(mode='r') as _:
                        return _.read()

                @property
                def json(self):
                    return json.loads(self.text)

            obj = Local(local_file, meta)

            try:
                # Keep reference to avoid id(...) clash
                _.downloads.append(obj)

                tcga.refs.annotations[obj] = meta
            except KeyError:
                raise RuntimeWarning(F"This error may happen if you download several file but discard the .now object.")

            return obj


download = _.Download()



# RA, 2020-06-17

import pandas
import tcga.refs


class _:
    dna0 = "ACGTWSMKRYBDHVN"
    dna1 = "TGCAWSKMYRVHDBN"
    rna0 = "ACGUWSMKRYBDHVN"
    rna1 = "UGCAWSKMYRVHDBN"

    meta = {
        'data': "2020-06-17",
        'source': "https://en.wikipedia.org/wiki/Complementarity_(molecular_biology)",
    }

    class Complementer:
        def __init__(self, p: str, q: str):
            self._reversed = False
            self._p = p
            self._q = q
            self.m = dict(zip(p, q))

        def __call__(self, s: str):
            return ''.join(self.m[c] for c in (s[::-1] if self._reversed else s))

        @property
        def reverse(self):
            other = _.Complementer(self._p, self._q)
            other._reversed = not self._reversed
            return other

        @property
        def legend(self):
            df = pandas.DataFrame(
                columns=[
                    "Symbol", "Description", "A", "C", "G", "T", "U",
                ],
                data=[
                    ['A', 'adenine', 1, 0, 0, 0, 0],
                    ['C', 'cytosine', 0, 1, 0, 0, 0],
                    ['G', 'guanine', 0, 0, 1, 0, 0],
                    ['T', 'thymine', 0, 0, 0, 1, 0],
                    ['U', 'uracil', 0, 0, 0, 0, 1],
                    ['W', 'weak', 1, 0, 0, 1, 1],
                    ['S', 'strong', 0, 1, 1, 0, 0],
                    ['M', 'amino', 1, 1, 0, 0, 0],
                    ['K', 'keto', 0, 0, 1, 1, 1],
                    ['R', 'purine', 1, 0, 1, 0, 0],
                    ['Y', 'pyrimidine', 0, 1, 0, 1, 1],
                    ['B', 'not A', 0, 1, 1, 1, 1],
                    ['D', 'not C', 1, 0, 1, 1, 1],
                    ['H', 'not G', 1, 1, 0, 1, 1],
                    ['V', 'not T/U', 1, 1, 1, 0, 0],
                    ['N', 'any base', 1, 1, 1, 1, 1],
                ]
            )

            return df


dna_to_dna = _.Complementer(_.dna0, _.dna1)
dna_to_rna = _.Complementer(_.dna0, _.rna1)
rna_to_rna = _.Complementer(_.rna0, _.rna1)
rna_to_dna = _.Complementer(_.rna0, _.dna1)

tcga.refs.annotations[dna_to_dna] = _.meta
tcga.refs.annotations[dna_to_rna] = _.meta
tcga.refs.annotations[rna_to_rna] = _.meta
tcga.refs.annotations[rna_to_dna] = _.meta

if __name__ == '__main__':
    print("Nucleotides:", dna_to_dna.legend, sep='\n')



# RA, 2020-06-18

import datetime


class Annotations:
    def __init__(self):
        self.d = {}
        self[self] = {
            'timestamp': datetime.datetime.now(tz=datetime.timezone.utc).isoformat(sep=' '),
        }

    def __setitem__(self, obj, value: dict):
        if type(value) is not dict:
            raise ValueError(F"'References' only accepts a value of type dict, got '{type(value)}'.")
        if id(obj) in self.d:
            raise KeyError(F"The key '{id(obj)}' of object already exists (obj: {obj}).")
        self.d[id(obj)] = dict(value)

    def __getitem__(self, obj):
        if obj not in self:
            self[obj] = dict()

        return self.d[id(obj)]

    def __contains__(self, obj):
        return (id(obj) in self.d)

    # def __call__(self, value: dict, obj):
    #     self[obj] = value
    #     return obj


annotations = Annotations()



# RA, 2020-06-17

import tcga.utils
import typing


def circular_windows(v: slice, n: int) -> typing.Iterable[slice]:
    """
    For an extended slice `v`, generate a set of slices
    for a sequence of length `n`, so that they
    together make up `v`.

    For example, if
        v = slice(-3, 5, 2)  and  n = 3
    then yield the slices
        [slice(0, 3, 2), slice(1, 3, 2), slice(0, 2, 2)]
    In particular, if
        s = 'ABC'
    then
        [s[w] for w in circular_windows(v, len(s))]
    is the list
        ['AC', 'B', 'A']
    because those compose the slice `v`:
       -3  01234
        * * * *
        ABCABCABC
    """

    (a, b, s) = (v.start, v.stop, v.step)

    assert isinstance(n, int), "The length of the sequence/string is an integer"
    assert (n >= 0), "The length of the sequence/string must be nonnegative"

    empty = slice(0, 0)

    if (n <= 0):
        yield empty
        return

    if (b <= a) and (s is None):
        yield empty
        return

    if (a <= b) and ((s is not None) and (s < 0)):
        yield empty
        return

    if (b <= a) and ((s is not None) and (s > 0)):
        yield empty
        return

    # Simple forward case
    if (a <= b) and ((s is None) or (s == 1)):
        d = b - a
        while d:
            a = a % n
            i = (min(n, a + d) - a)
            yield slice(a, a + i)
            d -= i
            a += i
        return

    # Forward case with positive step
    if (a <= b) and (s > 0):
        d = b - a
        a = a % n
        while d > 0:
            b = min(n, a + d)
            assert (0 <= a < b <= n)
            yield slice(a, b, s)
            if (b - a) % s:
                b += s - (b - a) % s
            d -= (b - a)
            a += (b - a) - n
        return

    # Backward case
    if (b < a) and (s < 0):
        d = b - a
        a = a % n
        while d < 0:
            b = max(-1, a + d)
            yield slice(a, max(b, 0), s)
            if (b == -1) and not (a % s):
                yield slice(0, 1)
            if (b - a) % s:
                b += s - (b - a) % s
            d -= (b - a)
            a += (b - a) + n
        return

    raise NotImplementedError


class _Laola:
    class Viewer:
        def __init__(self, key):
            self.key = key

        def __call__(self, x: typing.Union[str, list, tuple]):
            if isinstance(self.key, slice):
                self.key.indices(1)
                parts = (x[w] for w in circular_windows(self.key, len(x)))
                return tcga.utils.join[type(x)](parts)

            if isinstance(self.key, int):
                return x[self.key % len(x)]

            raise NotImplementedError

    def __getitem__(self, key):
        return _Laola.Viewer(key)


laola = _Laola()


class Circular:
    """
    Provides a view of a string/list/tuple
    pretending that it is circular.

    For example,
        Circular("ABC")[-3:5:2]
    returns
        "ACBA"
    because:
        ABCABCABC
       -3  012345
        * * * *

    The view returns a shallow copy.

    Note: len/str/repr refer to the original.
    """

    def __init__(self, x):
        self.x = x

    def __getitem__(self, key):
        return laola[key](self.x)

    def __len__(self):
        return len(self.x)

    def __repr__(self):
        return F"{type(self).__name__}({repr(self.x)})"


def _example():
    S = "a string"
    assert S[3:1] == ""

    for n in range(-10, 30):
        assert Circular(S)[n] == S[n % len(S)]

    assert Circular(S)[3:1] == S[3:1]

    #
    print("Example 1")

    w = laola[-3:5:2]
    s = "ABC"
    print(F"If w = laola[-3:5:2] and s = '{s}' then")
    print(F"    w(s) = '{w(s)}'")

    #
    print("Example 2")

    v = slice(-3, 5, 2)
    s = "ABC"
    print(F"If v = {v} and s = '{s}' then")
    print(F"    [s[w] for w in circular_windows(v, len(s))]")
    print(F"is")
    print(F"    {[s[w] for w in circular_windows(v, len(s))]}")

    #
    print(F"Examples of {Circular.__name__}:")

    for n in range(0, 30):
        a = -n // 2
        b = n
        s = Circular(S)
        print(F"{s}[{a}:{b}:2] is '{s[a:b:2]}'")


if __name__ == '__main__':
    _example()



# RA, 2020-06-17

"""
All members of this module must be annotated
in particular containing the field 'source'.
"""

import tcga.strings
import tcga.refs
import pathlib
import zipfile
import pickle
import json
import pandas

# On codes
# https://en.wikipedia.org/wiki/Translation_(biology)#Translation_tables

standard_rna = dict(tcga.strings.nnna(
    # (Phe/F) Phenylalanine
    'UUU F' 'UUC F'
    # (Leu/L) Leucine
    'UUA L' 'UUG L' 'CUU L' 'CUC L' 'CUA L' 'CUG L'
    # (Ile/I) Isoleucine
    'AUU I' 'AUC I' 'AUA I'
    # (Met/M) Methionine
    'AUG M'
    # (Val/V) Valine
    'GUU V' 'GUC V' 'GUA V' 'GUG V'
    # (Ser/S) Serine
    'UCU S' 'UCC S' 'UCA S' 'UCG S'
    # (Pro/P) Proline
    'CCU P' 'CCC P' 'CCA P' 'CCG P'
    # (Thr/T) Threonine
    'ACU T' 'ACC T' 'ACA T' 'ACG T'
    # (Ala/A) Alanine
    'GCU A' 'GCC A' 'GCA A' 'GCG A'
    # (Tyr/Y) Tyrosine
    'UAU Y' 'UAC Y'
    # Stop (Ochre, Amber, Opal)
    'UAA *' 'UAG *' 'UGA *'
    # (His/H) Histidine
    'CAU H' 'CAC H'
    # (Gln/Q) Glutamine
    'CAA Q' 'CAG Q'
    # (Asn/N) Asparagine
    'AAU N' 'AAC N'
    # (Lys/K) Lysine
    'AAA K' 'AAG K'
    # (Asp/D) Aspartic acid
    'GAU D' 'GAC D'
    # (Glu/E) Glutamic acid
    'GAA E' 'GAG E'
    # (Cys/C) Cysteine
    'UGU C' 'UGC C'
    # (Trp/W) Tryptophan
    'UGG W'
    # (Arg/R) Arginine
    'CGU R' 'CGC R' 'CGA R' 'CGG R' 'AGA R' 'AGG R'
    # (Ser/S) Serine
    'AGU S' 'AGC S'
    # (Gly/G) Glycine
    'GGU G' 'GGC G' 'GGA G' 'GGG G'
))

tcga.refs.annotations[standard_rna] = {
    'source': "https://en.wikipedia.org/wiki/Genetic_code#Standard_codon_tables",
    'date': "2020-06-17",
}


class _:
    # Load from this file
    parsed = (pathlib.Path(__file__).parent / "static/parsed/tables.pkl.zip")

    with zipfile.ZipFile(parsed, mode='r') as zf:
        with zf.open("data", mode='r') as fd:
            tables = pickle.load(fd)
        with zf.open("meta", mode='r') as fd:
            meta = json.load(fd)


tables: pandas.DataFrame
tables = _.tables.set_index('id')
tcga.refs.annotations[tables] = _.meta

assert hasattr(tables, 'name')
assert hasattr(tables, 'short_name')
assert hasattr(tables, 'dna_codons')
assert hasattr(tables, 'rna_codons')



# RA, 2020-06-17

import typing
import itertools


class _:
    class Join:
        def __init__(self, as_type=str):
            self._result_type = as_type

        def __getitem__(self, as_type=str):
            return _.Join(as_type)

        def __call__(self, iterable: typing.Iterable):
            if self._result_type is str:
                return ''.join(itertools.chain.from_iterable(iterable))
            else:
                return self._result_type(itertools.chain.from_iterable(iterable))


class First:
    """
    Example:
        f = First(list.pop).then(str.lower).then(tuple)
        assert f(["ABC", "XYZ"]) == ('x', 'y', 'z')
    """

    @classmethod
    def _as_callable(cls, f):
        if isinstance(f, typing.Callable):
            return f
        if hasattr(f, "__getitem__"):
            return f.__getitem__
        raise RuntimeError("It appears that the passed object f supports neither f(-) nor f[-].")

    def __init__(self, f):
        self.__ff = []
        self.then(f)

    def then(self, f):
        self.__ff.append(self._as_callable(f))
        return self

    def each(self, f):
        self.__ff.append(
            lambda c: map(self._as_callable(f), c)
        )
        return self

    def join(self, as_type):
        self.__ff.append(_.Join(as_type))
        return self

    def dict(self, keys) -> dict:
        """
        Convert to a dictionary on `keys`.
        """
        return {k: self(k) for k in keys}

    def __matmul__(self, other):
        return First(other).then(self)

    def __call__(self, x):
        for f in self.__ff:
            x = f(x)
        return x

    def __getitem__(self, x):
        return self(x)


join = _.Join()



# RA, 2020-06-18

import pandas
import pathlib
import tcga.refs.annotations

class _:
    datapath = pathlib.Path(__file__).parent / "static/data"
    file1 = datapath / "aa_prop1.csv"
    file2 = datapath / "aa_prop2.csv"
    nicer = (lambda s: ' '.join(s.split()))


properties = (
    pandas.read_csv(_.file1, sep='\t', comment='#')
).merge(
    pandas.read_csv(_.file2, sep='\t', comment='#').rename(columns={
        'pH 2': "At pH 2",
        'pH 7': "At pH 7",
    }),
    on='Abbr3',
)

tcga.refs.annotations[properties] = {
    'date': "2020-06-18",
    'source': [
        "https://www.sigmaaldrich.com/life-science/metabolomics/learning-center/amino-acid-reference-chart.html",
        "http://archive.ph/qL8ek",
    ],
    'comments': [
        _.nicer(
            """
            Columns 'At pH 2' and 'At pH 7' contain a measure of hydrophobicity.
            From the SA website: 
            The hydrophobicity index is a measure of the relative hydrophobicity, or how soluble an amino acid is in water. 
            In a protein, hydrophobic amino acids are likely to be found in the interior, 
            whereas hydrophilic amino acids are likely to be in contact with the aqueous environment.
            The values in the table are normalized so that the most hydrophobic residue 
            is given a value of 100 relative to glycine, which is considered neutral (0 value). 
            The scales were extrapolated to residues which are more hydrophilic than glycine.
            """
        ),
    ],
}

if __name__ == '__main__':
    print(properties)
    print(tcga.refs.annotations[properties])



import setuptools

# with open("README.md", 'r') as fd:
#     long_description = fd.read()

setuptools.setup(
    name="tcga",
    version="0.0.15",
    author="RA",
    author_email="numpde@null.net",
    keywords="computational biology bioinformatics genetics",
    description="Computational biology & bioinformatics utils",
    long_description="Computational biology & bioinformatics utils. Please visit the [github page](https://github.com/numpde/tcga).",
    long_description_content_type="text/markdown",
    url="https://github.com/numpde/tcga",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[''],

    # Required for includes in MANIFEST.in
    include_package_data=True,

    test_suite="nose.collector",
    tests_require=["nose"],
)



# RA, 2020-06-19

import pandas
import zipfile
import pickle
import tcga.refs
import pathlib
import json


class _:
    file = pathlib.Path(__file__).parent / "static/parsed/aaindex.pkl.zip"
    with zipfile.ZipFile(file, mode='r') as zf:
        with zf.open("data", mode='r') as fd:
            data = pickle.load(fd)
        with zf.open("meta", mode='r') as fd:
            meta = {'source': json.loads(fd.read().decode())}


data: pandas.Series
indices: pandas.Series
matrices: pandas.Series
potentials: pandas.Series

data = _.data['aaindex']
indices = _.data['list_of_indices']
matrices = _.data['list_of_matrices']
potentials = _.data['list_of_potentials']

tcga.refs.annotations[data] = _.meta
tcga.refs.annotations[indices] = _.meta
tcga.refs.annotations[matrices] = _.meta
tcga.refs.annotations[potentials] = _.meta

if __name__ == '__main__':
    print(data[indices.sample(1).index[0]].I)
    print(data['DOSZ010101'].M)
    print(_.meta)

del _
